import json
from arms.utils import exe


def build_system():
    # docker相关
    assert exe('apt-get update')
    assert exe('apt-get install -y apt-transport-https ca-certificates curl software-properties-common')
    assert exe('curl -fsSL https://download.daocloud.io/docker/linux/ubuntu/gpg | sudo apt-key add -')
    assert exe('add-apt-repository "deb [arch=$(dpkg --print-architecture)] https://download.daocloud.io/docker/linux/ubuntu $(lsb_release -cs) stable"')
    assert exe('apt-get update')
    assert exe('apt-get install -y -q docker-ce')
    assert exe('service docker start')
    assert exe('docker ps')
    open("/etc/docker/daemon.json", "w").write(json.dumps({
        "insecure-registries": ["rap.parsec.com.cn:9980"],
        "registry-mirrors": ["http://11854a36.m.daocloud.io"]
    }))
    assert exe('service docker restart')
    # docker-compose相关
    assert exe('wget http://h5.parsec.com.cn/arms/docker-compose-$(uname -s)-$(uname -m)')
    assert exe('chmod +x docker-compose-Linux-x86_64')
    assert exe('mv docker-compose-Linux-x86_64 /usr/local/bin/docker-compose')
    # 前端build工具
    assert exe('curl -sL https://deb.nodesource.com/setup_8.x | bash -')
    assert exe('curl -sS https://dl.yarnpkg.com/debian/pubkey.gpg | sudo apt-key add -')
    assert exe('echo "deb https://dl.yarnpkg.com/debian/ stable main" | sudo tee /etc/apt/sources.list.d/yarn.list')
    assert exe('apt-get update && sudo apt-get install -y nodejs yarn')
    # gitlab-runner相关
    # assert exe('curl -L https://packages.gitlab.com/install/repositories/runner/gitlab-runner/script.deb.sh | sudo bash')
    assert exe('curl https://packages.gitlab.com/gpg.key 2> /dev/null | sudo apt-key add - &>/dev/null')
    open('/etc/apt/sources.list.d/gitlab-runner.list', 'w').write(
        'deb https://mirrors.tuna.tsinghua.edu.cn/gitlab-runner/ubuntu xenial main\n')
    assert exe('apt-get update')
    assert exe('apt-get install -y gitlab-runner')
    assert exe('usermod -a -G docker gitlab-runner')
    assert exe('mkdir -p /opt/docker_volumes && chown gitlab-runner /opt/docker_volumes')
    assert exe('mkdir -p /mnt/docker_volumes && chown gitlab-runner /mnt/docker_volumes')
    assert exe('docker network create common')



import re


def prefix(lines, f):
    for line in lines:
        tmp = f(line)
        if tmp:
            yield tmp
        else:
            break


def extract_include(line):
    if not line: return None
    if line.startswith('`:include ') and line[-1] == '`':
        return line[1:-1].split(':include', 1)[-1].strip()
    else:
        return None


def lined_group(lines, starter, ender):
    def get(p):
        return lines[p] if p < len(lines) else None

    def single_group(p):
        next = ''
        isGrouping = False
        first = last = None
        for i in range(p, len(lines) + 1):
            prev, next = next, get(i)
            if not isGrouping and starter(prev, next):
                first, isGrouping = i, True
            elif isGrouping and ender(prev, next):
                last = i
                return (first, last)
        return (first, last)

    i = 0
    while True:
        first, last = single_group(i)
        if first is None or last is None:
            break
        yield lines[first:last]
        i = last


def mdtitle_level(line):
    """
    >>> mdtitle_level("#### hehe")
    4
    >>> mdtitle_level("#haha")
    1
    >>> mdtitle_level(".#haha")
    0
    """
    segs = re.findall("^#+", line)
    return len(segs[0]) if segs else 0


def big_mdtitle_group(lines):
    for group in lined_group(
            lines,
            lambda _, next: next is not None and mdtitle_level(next) == 1,
            lambda _, next: next is None or mdtitle_level(next) == 1):
        yield group


def small_mdtitle_group(lines):
    for group in lined_group(
            lines,
            lambda _, next: next is not None and mdtitle_level(next) >= 2,
            lambda _, next: next is None or mdtitle_level(next) >= 2):
        yield group


def mdyaml_group(lines):
    for group in lined_group(
            lines,
            lambda _, next: next is not None and next == '```yaml',
            lambda prev, next: next is None or prev == '```'):
        yield group


def pathparam_names(path):
    """
    >>> pathparam_names("/school/{schoolId}")
    ['schoolId']
    >>> pathparam_names("/{school}/{schoolId}")
    ['school', 'schoolId']
    >>> pathparam_names("/schools")
    []
    """
    return re.findall(r'\{([^/]+)\}', path)


def merge_dicts(dicts):
    """
    >>> merge_dicts([{'a':1, 'b':2}, {'c':3, 'b':4}])
    {'a': 1, 'b': 4, 'c': 3}
    """
    ret = {}
    for d in dicts:
        ret.update(d)
    return ret


def list_find(arr, f):
    for i, item in enumerate(arr):
        if f(item):
            return i
    return -1


def is_comment_format(value):
    """
    >>> is_comment_format("+int = 1 //数量")
    True
    >>> is_comment_format("abcde #hehe")
    False
    """
    if not isinstance(value, str) or '//' not in value:
        return False
    head = value.rsplit('//', 1)[0]
    if '=' not in head:
        return False
    pytype, sample = head.split('=', 1)
    if not sample.strip():
        return False
    return all(ord(c) < 128 for c in pytype.strip())


def parse_comment(value):
    """
    >>> parse_comment("+int = 1 //数量")
    ('+int', '1', '数量', '')
    >>> parse_comment('str ="abc"// 文本 #A')
    ('str', '"abc"', '文本', 'A')
    >>> parse_comment('int = 22 //hehe')
    ('int', '22', 'hehe', '')
    >>> parse_comment('str(50) =\"22\"//hehe')
    ('str(50)', '"22"', 'hehe', '')
    >>> parse_comment('  int=2 //#')
    ('int', '2', '', '')
    """
    assert is_comment_format(value), "invalid value: " + str(value)
    head, tail = value.rsplit('//', 1)
    pytype, sample = head.split('=', 1)
    comment, tag = tail.split('#', 1) if '#' in tail else (tail, '')
    return pytype.strip(), sample.strip(), comment.strip(), tag.strip()



import yaml
import os
from arms.yapi.util import parse_comment, list_find, is_comment_format


def java_type_of(pytype):
    typename_map = {
        'int': 'Integer',
        'long': 'Long',
        'bool': 'Boolean',
        'float': 'Double',
        'str': 'String',
        'text': 'String',
        'datetime': 'LocalDateTime',
        'date': 'LocalDate',
        'time': 'LocalTime',
        'json': 'String',
    }
    if pytype.startswith('+'):
        return '+' + java_type_of(pytype[1:])
    if pytype.startswith('str('):
        return 'String'
    return typename_map.get(pytype, None)


def split_jclass(jclass_text):
    lines = jclass_text.splitlines()
    p = list_find(lines, lambda x: ' class ' in x)
    q = list_find(lines, lambda x: x.strip() == '}')
    assert p >= 0
    return '\n'.join(lines[:p+1]), '\n'.join(lines[p+1:q]), '\n'.join(lines[q:])


def diff_jcode(old_jcode, new_jcode, filename):
    def parse(jcode):
        ret = set()
        for line in jcode.splitlines():
            line = line.strip()
            if not line or line.startswith('@') or line.startswith('//'): continue
            if '{' in line: break
            ret.add(line)
        return ret

    exit_code = 0
    old_keys = parse(old_jcode)
    new_keys = parse(new_jcode)
    for key in old_keys:
        if key not in new_keys:
            print('Need deleting: %s  (%s)' % (key, filename))
            exit_code = 1
        else:
            new_keys.remove(key)

    for key in new_keys:
        print('Need inserting: %s  (%s)' % (key, filename))
        exit_code = 1

    return exit_code


jline_tpl = '{local_head}{sql_head}{jtype} {name};  //{comment}'

jclass_tpl = """package {javaroot}.entity;
import com.parsec.universal.dao.annot.*;
import com.parsec.universal.dao.model.BaseModel;
import lombok.Data;
import org.springframework.format.annotation.DateTimeFormat;
import java.time.*;

{sql_head}
@Data
//{model_comment}
public class {model_name} extends BaseModel <(
{jcode}
)>
"""


def jcode_of(model_name, model_dict):
    jlines = []
    is_table = False
    model_comment = ''
    for name, value in model_dict.items():
        if name == '//':
            model_comment = value
            continue
        assert is_comment_format(value), "invalid value: " + str(value)
        pytype, realvalue, comment, _ = parse_comment(value)
        jtype = java_type_of(pytype)
        assert jtype is not None, "invalid type: " + value
        local_head = sql_head = ''
        if name.lower() == model_name.lower() + 'id':
            is_table = True
            assert not jtype.startswith('+'), "invalid private key: " + value
            sql_head = '@ParsecPrimaryKey\n'
        elif jtype.startswith('+'):
            jtype, sql_head = jtype[1:], '@ParsecNotDBColumn\n'

        if jtype == 'LocalDateTime':
            local_head = '@DateTimeFormat(pattern = "yyyy-MM-dd HH:mm:ss")\n'
        elif jtype == 'LocalDate':
            local_head = '@DateTimeFormat(pattern = "yyyy-MM-dd")\n'
        elif jtype == 'LocalTime':
            local_head = '@DateTimeFormat(pattern = "HH:mm:ss")\n'

        jline = jline_tpl.format(local_head=local_head, sql_head=sql_head, jtype=jtype, name=name, comment=comment)
        jlines.extend(jline.splitlines())

    jcode = '\n'.join('    ' + jline for jline in jlines)
    return dict(jcode=jcode, model_comment=model_comment, is_table=is_table)


def jcontent_of(model_name, model_dict, javaroot):
    """
    >>> output_dict = {'//': 'comment', 'aa': '+bool=1//', 'bb': 'datetime="true"//n#abc', 'cc': 'long=22 //hehe'}
    >>> print(jcontent_of('User', output_dict, 'com.corp.proj'))
    @ParsecNotDBColumn
    Boolean aa;  //
    @DateTimeFormat(pattern = "yyyy-MM-dd HH:mm:ss")
    LocalDateTime bb;  //n
    Long cc;  //hehe
    """
    jcode_ret = jcode_of(model_name, model_dict)
    jcode = jcode_ret['jcode']
    is_table = jcode_ret['is_table']
    model_comment = jcode_ret['model_comment']
    sql_head = '@ParsecTable("Tbl{}")'.format(model_name) if is_table else ''
    return jclass_tpl.format(javaroot=javaroot, sql_head=sql_head, model_name=model_name, jcode=jcode, model_comment=model_comment
                             ).replace('<(', '{').replace(')>', '}')


def process(infile, outpath, javaroot, checkonly):
    model_data = yaml.load(open(infile).read())
    for model_name, model_dict in model_data.items():
        filename = os.path.join(outpath, model_name + '.java')
        new_jcode = jcode_of(model_name, model_dict)['jcode']
        if not checkonly:
            if os.path.isfile(filename):
                head, jcode, tail = split_jclass(open(filename).read())
                with open(filename, 'w') as f:
                    f.write('{head}\n{jcode}\n{tail}\n'
                            .format(head=head, tail=tail, jcode=new_jcode))
            else:
                with open(filename, 'w') as f:
                    f.write(jcontent_of(model_name, model_dict, javaroot))
        else:
            if os.path.isfile(filename):
                head, jcode, tail = split_jclass(open(filename).read())
                exit_code = diff_jcode(jcode, new_jcode, model_name + '.java')
            else:
                print('Need Model File: ' + filename)
                exit_code = 1
            if exit_code:
                exit(exit_code)




# coding: utf-8
"""
arms
~~~~~~~~
CI/CD tool of Chongqing Parsec Corp.
Setup
-----
.. code-block:: bash
    > pip install arms
    > arms -h

"""

from setuptools import setup, find_packages  # Always prefer setuptools over distutils
from codecs import open  # To use a consistent encoding
from os import path
from setuptools.command.install import install
import re
import ast

_version_re = re.compile(r'__version__\s+=\s+(.*)')
version = str(ast.literal_eval(
    _version_re.search(
        open('arms/__init__.py').read()
    ).group(1)
))
here = path.abspath(path.dirname(__file__))


class MyInstall(install):
    def run(self):
        print("-- installing... --")
        install.run(self)

setup(
        name = 'arms',
        version=version,
        description='CI/CD tool of Chongqing Parsec Corp.',
        long_description='\npip install arms\n\n'
                         'arms -h',
        url='https://pypi.python.org/pypi/arms',
        author='qorzj',
        author_email='inull@qq.com',
        license='MIT',
        platforms=['any'],

        classifiers=[
            ],
        keywords='arms arms armstrong',
        packages = ['arms', 'arms.yapi', 'arms.yapi.java_util'],
        install_requires=['lesscli', 'requests', 'Jinja2', 'PyYAML==3.12', 'pyperclip'],

        cmdclass={'install': MyInstall},
        entry_points={
            'console_scripts': [
                'arms = arms.main:entrypoint'
                ],
            },
    )
